#include "die.h"

#include <stdarg.h>
#include <stdlib.h>
#include "ev.h"
#include "evnames.h"
#include "analyze_output.h"
#include "msgconfig.h"

static void printEvent(const EventRecord *ep) {
	fprintf(fp_output,
	        "thread-id: %ld\n"
	        "timestamp= %17.6f\n"
	        "mtx-addr = %08lx\n"
	        "cond-addr = %08lx\n"
	        "event = %s\n",
	        ep->tid,
	        ep->timestamp,
	        ep->mtx_addr,
	        ep->cond_addr,
	        ev_name[ep->event]
	       );
}

void die(const EventRecord *ep, int msg, const char *fmt, ...) {
	OutputParagraphStart();
	if(ep) {
		fprintf(fp_output,"Fatal error found in event\n");
		printEvent(ep);
	}
	fprintf(fp_output,"Error (message code %d): ",msg);
	va_list ap;
	va_start(ap,fmt);
	vfprintf(fp_output,fmt,ap);
	va_end(ap);
	OutputParagraphEnd();
	fflush(fp_output);
	exit(EXIT_FAILURE);
}

void warn(const EventRecord *ep, int msg, const char *fmt, ...) {
	if(!isMessageEnabled(msg)) return;
	OutputParagraphStart();
	if(ep) {
		fprintf(fp_output,"Warning found in event\n");
		printEvent(ep);
	}
	fprintf(fp_output,"Warning (message code %d): ",msg);
	va_list ap;
	va_start(ap,fmt);
	vfprintf(fp_output,fmt,ap);
	va_end(ap);
	OutputParagraphEnd();
}

void suggest(const EventRecord *ep, int msg, const char *fmt, ...) {
	if(!isMessageEnabled(msg)) return;
	OutputParagraphStart();
	if(ep) {
		fprintf(fp_output,"Suggestion for event\n");
		printEvent(ep);
	}
	fprintf(fp_output,"Suggestion (message code %d): ",msg);
	va_list ap;
	va_start(ap,fmt);
	vfprintf(fp_output,fmt,ap);
	va_end(ap);
	OutputParagraphEnd();
}
