#include <pthread.h>
#include <dlfcn.h>
#include <unistd.h>
#include <fcntl.h>
#include <stdio.h> /*sprintf only*/
#include <sys/time.h>
#include <string.h>

#include "ev.h"
#include "evnames.h"

#ifdef __HPUX__
/*HP/UX is just different...*/
#define PTHREAD_CREATE __pthread_create_system
#define PTHREAD_CREATE_NAME "__pthread_create_system"
#define PTHREAD_LIB_NAME "libpthread.1"
#endif
#ifndef PTHREAD_CREATE
#define PTHREAD_CREATE pthread_create
#define PTHREAD_CREATE_NAME "pthread_create"
#define PTHREAD_LIB_NAME "libpthread.so"
#endif


#ifdef POINTERSARE64BIT
#define PTRFORMAT "016lx"
#else
#define PTRFORMAT "08lx"
#endif

#ifdef __LINUX__
#ifdef __cplusplus
#define NOTHROW throw()
#else
#define NOTHROW
#endif
#else
#define NOTHROW
#endif

#ifdef __cplusplus
extern "C" {
#endif

static void *lib_pthread=0;
static int eventlog_fd=-1;
static int initializing=0;

static union {
	void *ptr00;
	pthread_t (*pthread_self_ptr)();
} u00;

static union {
	void *ptr0;
	int (*pthread_mutex_init_ptr)(pthread_mutex_t*, const pthread_mutexattr_t*);
} u0;
static union {
	void *ptr1;
	int (*pthread_mutex_lock_ptr)(pthread_mutex_t*);
} u1;
static union {
	void *ptr2;
	int (*pthread_mutex_unlock_ptr)(pthread_mutex_t*);
} u2;
static union {
	void *ptr3;
	int (*pthread_mutex_trylock_ptr)(pthread_mutex_t*);
} u3;
static union {
	void *ptr4;
	int (*pthread_mutex_destroy_ptr)(pthread_mutex_t*);
} u4;
static union {
	void *ptr10;
	int (*pthread_create_ptr)(pthread_t*,const pthread_attr_t*,void*(*)(void*),void *);
} u10;
static union {
	void *ptr11;
/*GCC 2.95.3 and 3.0 disagree on this
 *#ifdef __LINUX__
 *	void (*pthread_exit_ptr)(void*) __attribute__ ((__noreturn__));
 *#else
 */
	void (*pthread_exit_ptr)(void*);
/*#endif*/
} u11;
static union {
	void *ptr12;
	int (*pthread_join_ptr)(pthread_t,void **);
} u12;
static union {
	void *ptr13;
	int (*pthread_detach_ptr)(pthread_t);
} u13;


static union {
	void *ptr20;
	int   (*pthread_cond_broadcast_ptr)(pthread_cond_t *);
} u20;
static union {
	void *ptr21;
	int   (*pthread_cond_destroy_ptr)(pthread_cond_t *);
} u21;
static union {
	void *ptr22;
	int   (*pthread_cond_init_ptr)(pthread_cond_t *, const pthread_condattr_t *);
} u22;
static union {
	void *ptr23;
	int   (*pthread_cond_signal_ptr)(pthread_cond_t *);
} u23;
static union {
	void *ptr24;
	int   (*pthread_cond_timedwait_ptr)(pthread_cond_t *, pthread_mutex_t *, const struct timespec *);
} u24;
static union {
	void *ptr25;
	int   (*pthread_cond_wait_ptr)(pthread_cond_t *, pthread_mutex_t *);
} u25;

static void initialize() {
	if(initializing) return;
	initializing=1;
	lib_pthread = dlopen(PTHREAD_LIB_NAME, RTLD_NOW);
	if(!lib_pthread) {
		const char *e = dlerror();
		write(2,"mta interposer could not load libpthread.1\n",43);
		write(2,e,strlen(e));
		write(2,"\n",1);
		_exit(1);
	}
	u00.ptr00 = (void*)dlsym(lib_pthread,"pthread_self");
	u0.ptr0 = (void*)dlsym(lib_pthread,"pthread_mutex_init");
	u1.ptr1 = (void*)dlsym(lib_pthread,"pthread_mutex_lock");
	u2.ptr2 = (void*)dlsym(lib_pthread,"pthread_mutex_unlock");
	u3.ptr3 = (void*)dlsym(lib_pthread,"pthread_mutex_trylock");
	u4.ptr4 = (void*)dlsym(lib_pthread,"pthread_mutex_destroy");

	u10.ptr10 = (void*)dlsym(lib_pthread,PTHREAD_CREATE_NAME);
	u11.ptr11 = (void*)dlsym(lib_pthread,"pthread_exit");
	u12.ptr12 = (void*)dlsym(lib_pthread,"pthread_join");
	u13.ptr13 = (void*)dlsym(lib_pthread,"pthread_detach");
	
	u20.ptr20 = (void*)dlsym(lib_pthread,"pthread_cond_broadcast");
	u21.ptr21 = (void*)dlsym(lib_pthread,"pthread_cond_destroy");
	u22.ptr22 = (void*)dlsym(lib_pthread,"pthread_cond_init");
	u23.ptr23 = (void*)dlsym(lib_pthread,"pthread_cond_signal");
	u24.ptr24 = (void*)dlsym(lib_pthread,"pthread_cond_timedwait");
	u25.ptr25 = (void*)dlsym(lib_pthread,"pthread_cond_wait");
	
	eventlog_fd = open("mta.log", O_WRONLY|O_CREAT|O_TRUNC|O_APPEND, 0644);
	if(eventlog_fd<0) {
		write(2,"mta interposer could not create log file\n",41);
		_exit(1);
	}
	initializing=0;
}

static void logEvent(event_t ev, pthread_mutex_t *pmtx, pthread_cond_t *pcond, long info) {
	char buf[128];
	struct timeval tv_now;
	pthread_t tid;
	gettimeofday(&tv_now,0);
	tid=(*u00.pthread_self_ptr)();
	sprintf(buf,"%"PTRFORMAT" %10ld.%06ld %2d %s %"PTRFORMAT" %"PTRFORMAT" %ld\n",(long)tid,tv_now.tv_sec,tv_now.tv_usec,ev,ev_name[ev],(long)pmtx,(long)pcond,info);
	write(eventlog_fd,buf,strlen(buf));
}

int pthread_mutex_init(pthread_mutex_t *pmtx, const pthread_mutexattr_t *pattr) NOTHROW {
	int rc;
	if(!lib_pthread) initialize();
	logEvent(ev_mutex_create,pmtx,0,0);
	rc = (*u0.pthread_mutex_init_ptr)(pmtx,pattr);
	logEvent(ev_mutex_created,pmtx,0,(long)rc);
	return rc;
}

int pthread_mutex_lock(pthread_mutex_t *pmtx) NOTHROW {
	int rc;
	if(!lib_pthread) initialize();
	if(initializing) return 0;
	logEvent(ev_mutex_lock,pmtx,0,0);
	rc = (*u1.pthread_mutex_lock_ptr)(pmtx);
	logEvent(ev_mutex_locked,pmtx,0,(long)rc);
	return rc;
}

int pthread_mutex_trylock(pthread_mutex_t *pmtx) NOTHROW {
	int rc;
	if(!lib_pthread) initialize();
	logEvent(ev_mutex_trylock,pmtx,0,0);
	rc = (*u3.pthread_mutex_trylock_ptr)(pmtx);
	if(rc==0)
		logEvent(ev_mutex_locked,pmtx,0,0);
	else
		logEvent(ev_mutex_trylock_failed,pmtx,0,(long)rc);
	return rc;
}

int pthread_mutex_unlock(pthread_mutex_t *pmtx) NOTHROW {
	int rc;
	if(initializing) return 0;
	logEvent(ev_mutex_unlock,pmtx,0,0);
	rc = (*u2.pthread_mutex_unlock_ptr)(pmtx);
	logEvent(ev_mutex_unlocked,pmtx,0,(long)rc);
	return rc;
}

int pthread_mutex_destroy(pthread_mutex_t *pmtx) NOTHROW {
	int rc;
	logEvent(ev_mutex_destroy,pmtx,0,0);
	rc = (*u4.pthread_mutex_destroy_ptr)(pmtx);
	logEvent(ev_mutex_destroyed,pmtx,0,(long)rc);
	return rc;
}


/* threads ------------------------------------------------------------------*/
int PTHREAD_CREATE(pthread_t *thread, const pthread_attr_t *attr,
                   void * (*start_routine)(void *), void * arg) NOTHROW 
{
	int rc;
	if(!lib_pthread) initialize();
	logEvent(ev_thread_create,0,0,0);
	rc = (*u10.pthread_create_ptr)(thread,attr,start_routine,arg);
	if(rc==0)
		logEvent(ev_thread_created,0,0,(long)*thread);
	else
		logEvent(ev_thread_create_failed,0,0,(long)rc);
	return rc;
}

void pthread_exit(void *retval) NOTHROW {
	if(!lib_pthread) initialize();
	logEvent(ev_thread_exit,0,0,(long)retval);
	(*u11.pthread_exit_ptr)(retval);
}

int pthread_join(pthread_t tid, void **return_value) NOTHROW {
	int rc;
	logEvent(ev_thread_join,0,0,(long)tid);
	rc = (*u12.pthread_join_ptr)(tid,return_value);
	logEvent(ev_thread_joined,0,0,(long)rc);
	return rc;
}

int pthread_detach(pthread_t tid) NOTHROW {
	int rc;
	logEvent(ev_thread_detach,0,0,(long)tid);
	rc = (*u13.pthread_detach_ptr)(tid);
	logEvent(ev_thread_detached,0,0,(long)rc);
	return rc;
}



/* conditions ---------------------------------------------------------------*/
int pthread_cond_broadcast(pthread_cond_t *pcond) NOTHROW {
	int rc;
	if(!lib_pthread) initialize();
	if(initializing) return 0;
	logEvent(ev_cond_broadcast,0,pcond,0);
	rc = (*u20.pthread_cond_broadcast_ptr)(pcond);
	logEvent(ev_cond_broadcasted,0,pcond,(long)rc);
	return rc;
}

int pthread_cond_destroy(pthread_cond_t *pcond) NOTHROW {
	int rc;
	if(!lib_pthread) initialize();
	if(initializing) return 0;
	logEvent(ev_cond_destroy,0,pcond,0);
	rc = (*u21.pthread_cond_destroy_ptr)(pcond);
	logEvent(ev_cond_destroyed,0,pcond,(long)rc);
	return rc;
}

int pthread_cond_init(pthread_cond_t *pcond, const pthread_condattr_t *pattr) NOTHROW {
	int rc;
	if(!lib_pthread) initialize();
	if(initializing) return 0;
	logEvent(ev_cond_create,0,pcond,0);
	rc = (*u22.pthread_cond_init_ptr)(pcond,pattr);
	if(rc==0)
		logEvent(ev_cond_created,0,pcond,0);
	else
		logEvent(ev_cond_create_failed,0,pcond,(long)rc);
	return rc;
}

int pthread_cond_signal(pthread_cond_t *pcond) NOTHROW {
	int rc;
	if(!lib_pthread) initialize();
	if(initializing) return 0;
	logEvent(ev_cond_signal,0,pcond,0);
	rc = (*u23.pthread_cond_signal_ptr)(pcond);
	logEvent(ev_cond_signalled,0,pcond,(long)rc);
	return rc;
}

int pthread_cond_timedwait(pthread_cond_t *pcond, pthread_mutex_t *pmtx, const struct timespec *pts) NOTHROW {
	int rc;
	logEvent(ev_cond_timedwait,pmtx,pcond,0);
	rc = (*u24.pthread_cond_timedwait_ptr)(pcond,pmtx,pts);
	if(rc==0)
		logEvent(ev_cond_waited,pmtx,pcond,0);
	else
		logEvent(ev_cond_timedwait_failed,pmtx,pcond,(long)rc);
	return rc;
}

int pthread_cond_wait(pthread_cond_t *pcond, pthread_mutex_t *pmtx) NOTHROW {
	int rc;
	logEvent(ev_cond_wait,pmtx,pcond,0);
	rc = (*u25.pthread_cond_wait_ptr)(pcond,pmtx);
	logEvent(ev_cond_waited,pmtx,pcond,(long)rc);
	return rc;
}

#ifdef __cplusplus
} /*extern C*/
#endif
